{==============================================================================|
| Project : Delphree - WebProvider                               | 001.000.000 |
|==============================================================================|
| Content: WebProvider Application MainCode                                    |
|==============================================================================|
| The contents of this file are subject to the Mozilla Public License Ver. 1.0 |
| (the "License"); you may not use this file except in compliance with the     |
| License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ |
|                                                                              |
| Software distributed under the License is distributed on an "AS IS" basis,   |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for |
| the specific language governing rights and limitations under the License.    |
|==============================================================================|
| The Original Code is WebProvider for Delphi Library.                         |
|==============================================================================|
| WebProvider Application MainCode developed by Radek Voltr (Czech Republic)   |
| WebProvider Application MainCode is Copyright (C) 1998, 1999.                |
| All Rights Reserved.                                                         |
|==============================================================================|
| Contributor(s):                                                              |
|==============================================================================|
| History: 10.2.1999 Version 001.000.000                                       |
|          First public release.                                               |
|==============================================================================}

unit WPApp;

interface

uses Windows, Classes, Dialogs, Sysutils, DBTables, Forms;

type
  { component to global extension setting}
  TWPGlobal = class(TComponent)
  private
    { Private Fields of TWPGlobal }
    FDescription: string;
    FCodeOverSeer: boolean;
    FConvertFieldsValue: boolean;
    procedure SetDescription(Val: string);
    procedure SetCodeOverSeer(const Value: boolean);
    procedure SetConvertFieldsValue(const Value: boolean);
  protected
    { Protected methods of TWPGlobal }
  public
    { Public methods of TWPGlobal }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published properties of TWPGlobal }
    property Description: string read FDescription write SetDescription;
    property CodeOverSeerDebug: boolean read FCodeOverSeer write SetCodeOverSeer;
    property ConvertFieldsValue: boolean read FConvertFieldsValue write SetConvertFieldsValue;
  end;

  {onWPAction - HTML generate event}
  TOnWPAction = procedure(var Response: string; Continue: boolean) of object;

  {WPCustomAction - every WPAction must implement Execute method}
  TWPCustomAction = class(TComponent)
  public
    function Execute(var Body: string): Boolean; virtual;
  end;

  {WPAction - do Action when path info or field info is o.k}
  TWPAction = class(TWPCustomAction)
  private
    { Private Fields of TWPAction }
    FPath: string;
    FEnabled: Boolean;
    FResponse: string;
//    FFieldName: string;
//    FFieldValue: string;
    FOnAction: TOnWPAction;
    FClearPrev: boolean;
    procedure SetPath(const Value: string);
    procedure SetClearPrev(const Value: boolean);
  protected
    { Protected methods of TWPAction }
    procedure Loaded; override;
  public
    { Public methods of TWPAction }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute(var Body: string): Boolean; override;
  published
    { Published properties of TWPAction }
    property Path: string read FPath write SetPath;
    // path is part of request > http://xxx.xxx.xx/yyy.exe/zzzz
    // /zzzz is Path
    // Action is triggered when Path is in request
    property Enabled: Boolean read FEnabled write FEnabled;
//    property FieldName: string read FFieldName write FFieldName;
//    property FieldValue: string read FFieldValue write FFieldValue;
    // Action is triggered when FieldName has FieldValue is in request

    // if Path is '' and FieldName='' then Action is triggered everytime

    property OnAction: TOnWPAction read FOnAction write FOnAction;

    property ClearPrev: boolean read FClearPrev write SetClearPrev;
    // if you set this property your HTTP output will clear previous outputs

  end;

  TWProviderApp = class(TComponent)
  private
    FResponse: string;
    // cgi
    fOut: THandle;
    fIn: THandle;
    FFields: TStrings;
    procedure Write(Response: string);
    procedure SetResponse(const Value: string);
    procedure SetFields(const Value: TStrings);
  public
  // private
    FMainModule: TDataModule;
    fTranslate, fDebug: boolean;

    // query field by name
    function FieldByName(Name: string): string;
    function Path: string;
    function ServerVariable(VarName: string): string;
    function ReplaceText(Data, Find, Replace: string): string;
    function DecodeText(Data: string): string;
    procedure GetHttpQueryFields;
// published
    procedure Initialize;
    procedure CreateForm(InstanceClass: TComponentClass; var Reference);
    procedure Run;
    constructor Create(AOwner: TComponent); override;
    property Response: string read FResponse write SetResponse;
    property Fields: TStrings read FFields write SetFields;
  end;

var
  Application: TWProviderApp = nil;

procedure Register;

implementation

{ WProviderApp }

uses uDbg, uDbgIntf;

var
  Debug: TNxDebugger;

procedure Register;
begin
  {Register component and add to WebProvider page in VCL.}
  RegisterComponents('WebProvider', [TWPGlobal]);
  RegisterComponents('WebProvider', [TWPAction]);
end;

constructor TWPGlobal.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {Add any other initialization code here.}
end;

destructor TWPGlobal.Destroy;
begin
  {Add any cleanup code here.}
  inherited Destroy;
end;

procedure TWPGlobal.SetCodeOverSeer(const Value: boolean);
begin
  FCodeOverSeer := Value;
end;

procedure TWPGlobal.SetConvertFieldsValue(const Value: boolean);
begin
  FConvertFieldsValue := Value;
end;

procedure TWPGlobal.SetDescription(Val: string);
begin
  if FDescription <> Val then
  begin
    FDescription := Val;
  end;
end;

function TWPCustomAction.Execute(var Body: string): Boolean;
begin
// Continue with processing
  Result := false;
end;

constructor TWPAction.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {Add any other initialization code here.}
end;

destructor TWPAction.Destroy;
begin
  {Add any cleanup code here.}
  inherited Destroy;
end;

procedure TWPAction.Loaded;
begin
  inherited Loaded;
  { component specific initialization code, etc. }
end;

procedure TWPAction.SetPath(const Value: string);
begin
  FPath := UpperCase(Value);
  if Value <> '' then
    if (Value[1] <> '/') and (Value[1] <> '\') then FPath := '/' + UpperCase(Value);
end;

function TWPAction.Execute;
var
  PathGo, FieldGo: boolean;
  AppPath: string;
begin
//
  Result := false;
  if not Enabled then exit;
  FieldGo := false;
  PathGo := false;
  if Length(Path) > 1 then
  begin
    AppPath := Application.Path;
    PathGo := Copy(Path, 2, Length(Path)) = Copy(AppPath, 2, Length(AppPath));
  end;

  if PathGo or FieldGo then
// execute action
  begin
    if Assigned(FOnAction) then
      FOnAction(fResponse, Result);
    if FClearPrev then
      Body := FResponse
    else
      Body := body + FResponse;
  end;
end;

procedure TWPAction.SetClearPrev(const Value: boolean);
begin
  FClearPrev := Value;
end;


constructor TWProviderApp.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fFields := TStringList.Create;
end;

procedure TWProviderApp.CreateForm(InstanceClass: TComponentClass;
  var Reference);
var
  I: Integer;
begin
  if FMainModule = nil then
    TComponent(FMainModule) := InstanceClass.Create(Self)
  else raise Exception.Create('Je povolen pouze jeden modul datasetu vytvareny pri startu');

  for I := 0 to Application.fmainmodule.ComponentCount - 1 do
    if Application.fmainmodule.Components[I] is TWpglobal then
      fDebug := (Application.fmainmodule.Components[I] as TWpglobal).CodeOverSeerDebug;
  for I := 0 to Application.fmainmodule.ComponentCount - 1 do
    if Application.fmainmodule.Components[I] is TWpglobal then
      fTranslate := (Application.fmainmodule.Components[I] as TWpglobal).ConvertFieldsValue;

  if fDebug then
  begin
    Debug := TNxDebugger.Create('Application', 'WPApplication main code');
    NxStartDebug;
  end;
end;

function TWProviderApp.Path: string;
var
  S: string;
begin
  S := UpperCase(ServerVariable('PATH_INFO'));
  if (Pos('/', S) <> 0) and (Pos('\', S) <> 0) then exit;
  S := Copy(S, 1, Length(S));
  if Pos('?', S) <> 0 then S := Copy(S, 1, Pos('?', S) - 1);
  if Pos('&', S) <> 0 then S := Copy(S, 1, Pos('&', S) - 1);
  Result := S;
end;

function TWProviderApp.FieldByName(Name: string): string;
begin
     Result:=FFields.Values[UpperCase(Name)];
end;

procedure TWProviderApp.Initialize;
begin
  Application := TWProviderApp.Create(nil);
end;

procedure TWProviderApp.Run;
var
  HTTPHeader: string;
  HTMLBody: string;
  I: Integer;
begin
  fIn := GetStdHandle(STD_INPUT_HANDLE);
  fOut := GetStdHandle(STD_OUTPUT_HANDLE);

  if Debug <> nil then Debug.SendMsg('TWPApplication RUN');

  fDebug := false;
  if LowerCase(ParamStr(1)) = 'debug' then
  begin
    fDebug := true;
    if Debug <> nil then Debug.SendMsg('Debug mode enabled from commmand line');
    if debug <> nil then Debug.SendMsg('TWPApplication will simulate GET');
  end;
  // fill query fields
  GetHttpQueryFields;

  try

    HTTPHeader := 'HTTP/1.0 200 OK' + #13 + #10 +
      'Content-Type: text/html' + #13 + #10 + #13 + #10;
    Write(HTTPHeader);

    HTMLBody := '';
    for I := 0 to Application.fmainmodule.ComponentCount - 1 do
      if Application.fmainmodule.Components[I] is TWPCustomAction then
        if (Application.fmainmodule.Components[I] as TWPCustomAction).Execute(HTMLBody) then break;

    Write(HTMLBody);
  except
  end;

end;
{ 'GET' implementation }

procedure TWProviderApp.SetResponse(const Value: string);
begin
  FResponse := Value;
end;

function TWProviderApp.ServerVariable(VarName: string): string;
var
  X: array[0..255] of char;
  XLen: DWord;
begin
// get server variable
{debug}
  if fDebug and (VarName = 'QUERY_STRING') then begin Result := ParamStr(3); exit; end;
  if fDebug and (VarName = 'PATH_INFO') then begin Result := ParamStr(2); exit; end;
  if fDebug and (VarName = 'REQUEST_METHOD') then begin Result := 'GET'; exit; end;

{CGI}
  XLen := 255;
  if GetEnvironmentVariable(PChar(VarName), X, XLen) > 0 then Result := StrPas(X) else Result := '';
end;

procedure TWProviderApp.Write(Response: string);
var
  Wr: DWord;
  X: PChar;
begin
// write response
  try
//GetMem(X,SizeOf(response));
    X := StrAlloc(SizeOf(response));
    StrPCopy(X, response);
    WriteFile(fOut, X[0], Length(Response), Wr, nil);
    if wr <> Length(Response) then
   // Exception
    begin
      if Debug <> nil then Debug.SendString('Error in string response write', Response);
      Exception.Create('Error in string response write');
    end;
    if Debug <> nil then Debug.SendString('Write string response', Response);
  finally
//FreeMem(X,Length(response));
    StrDispose(X);
  end;
end;

procedure TWProviderApp.SetFields(const Value: TStrings);
begin
  FFields.assign(Value);
end;

procedure TWProviderApp.GetHttpQueryFields;
var
  S, QueryString: string;
  Name, Value: string;
  QueryLen, bR: DWord;
  Buffer: PChar;
begin
  FFields.Clear;
  try
    if ServerVariable('REQUEST_METHOD') = 'GET' then
      QueryString := ServerVariable('QUERY_STRING')
    else
      if ServerVariable('REQUEST_METHOD') = 'POST' then
      begin
        QueryLen := StrToInt(ServerVariable('CONTENT_LENGTH'));
        Buffer := StrAlloc(QueryLen);
        FillChar(Buffer[0], QueryLen, #0);
        try
          ReadFile(fIn, Buffer[0], QueryLen, bR, nil);
          if QueryLen <> bR then exit;
          QueryString:=Copy(StrPas(Buffer),1,QueryLen);
        finally
          StrDispose(Buffer);
        end;
      end
      else
        exit;
  except
  end;

  QueryString := QueryString + '?';
  QueryString := ReplaceText(QueryString, '&', '?');
  repeat
    if Pos('?', QueryString) <> 0 then S := Copy(QueryString, 1, Pos('?', QueryString) - 1);
    QueryString := Copy(QueryString, Pos('?', QueryString) + 1, Length(QueryString));
    if (S = '') and (QueryString <> '') then Continue;
    if Pos('=', S) <> 0 then
    begin
     // fill values with uppercase names
      Name := UpperCase(Copy(S, 1, Pos('=', S) - 1));
      if fTranslate then
        Value := DecodeText(Copy(S, Pos('=', S) + 1, Length(S)))
      else
        Value := Copy(S, Pos('=', S) + 1, Length(S));
      FFields.Add(Name + '=' + Value);
    end;
  until QueryString = '';
end;

function TWProviderApp.ReplaceText(Data, Find, Replace: string): string;
var
  DataU, FindU: string;
begin
  Result := Data;
  FindU := UpperCase(Find);
  repeat
    Data := Result;
    DataU := UpperCase(Data);
    if Pos(FindU, DataU) = 0 then Result := Data
    else
      Result := Copy(Data, 1, Pos(FindU, DataU) - 1) + Replace + Copy(Data, Pos(FindU, DataU) + Length(FindU), Length(Data));
  until Pos(FindU, UpperCase(Result)) = 0;
end;

function TWProviderApp.DecodeText(Data: string): string;
var
  I: Integer;
  P: Boolean;
  SS: string;
begin
// decode Text
  Result := ReplaceText(Data, '+', ' ');
  P := False;
  SS := Result;
  Result := '';
  if Length(SS) < 1 then exit;
  repeat
    if SS[1] <> '%' then
    begin
      Result := Result + SS[1];
      SS := Copy(SS, 2, Length(SS));
    end
    else
    begin
      SS := Copy(SS, 2, Length(SS));
      Result := Result + Chr(StrToInt('$' + Copy(SS, 1, 2)));
      SS := Copy(SS, 3, Length(SS));
    end;
  until SS = '';
end;

initialization
  Debug := nil;

finalization
  try
    if Debug <> nil then Debug.Free;
  except
  end;

end.

